!******************************************************************************
!
! File : jd_support.f90
!   by : Alan Tackett
!   on : 02/09/00
!  for : Support routines for the jd_solver module
!
!
!******************************************************************************

!******************************************************************************
!  Calculates Aq and Bq
!******************************************************************************
Subroutine ABMUL(n, x, Hx, Ox)
  Use jd_misc
  Use hamiltonian

  Implicit NONE!!!

  Integer :: n
  Complex :: x(n)
  Complex :: Hx(n)
  Complex :: Ox(n)

  Complex :: PDOT(PLM_Max)
  Logical :: PDOT_Stored

  PDOT_Stored = .FALSE.

  Call CalcHxAndOx( x(1:n), Hx(1:n), Ox(1:n), &
       PDOT, Mem_Kpnt, PDOT_Stored)

  Return
End Subroutine

!******************************************************************************
!  q = K^(-1) q
!******************************************************************************
Subroutine PreCon(n, q)
  Use jd_misc
  Use relax
  Use hamiltonian

  Implicit NONE!!!

  Integer :: n
  Complex :: q(n)
  Real    :: KE

  Integer :: gi, DN
  Real    :: G(3), Kvec(3), G2
  Complex, Pointer :: Diag(:)

RETURN


  Call GetBuffer(Diag)
  Call JD_Precond(Diag, Q(1:N), Mem_Kpnt)
  Q(1:N) = diag
  Call FreeBuffer(Diag)
RETURN

  KE = CalcKineticenergy(BZ%Ku(1:3,Mem_Kpnt), Q(1:N))

  Call GetBuffer( Diag)
  Call KE_Precond(Diag, KE, BZ%Ku(1:3,Mem_Kpnt))


!  Diag(1:Gpnt_Size(G_LOW)) = -Gpnt(4, 1:Gpnt_Size(G_Low)) 
!  Diag(Gpnt_Size(G_Low)+1:) = -Gpnt(4, 2:Gpnt_Size(G_LOW))
!  Diag = Diag + SCFvalues%Ve_fourier(1) - Shift_Energy 

  KE = 1.0
  Diag(1:Gpnt_Size(G_LOW)) = -KE*Gpnt(4, 1:Gpnt_Size(G_Low)) + &
       SCFvalues%Ve_fourier(1:Gpnt_Size(G_Low)) - Shift_Energy
  Diag(Gpnt_Size(G_Low)+1:) = -KE*Gpnt(4, 2:Gpnt_Size(G_LOW)) + &
       CONJG(SCFvalues%Ve_fourier(2:Gpnt_Size(G_Low))) - Shift_Energy

!  Diag(1:Gpnt_Size(G_LOW)) = 1.0/Diag(1:Gpnt_Size(G_Low)) + &
!       SCFvalues%Ve_fourier(1:Gpnt_Size(G_Low)) - Shift_Energy
!  Diag(Gpnt_Size(G_Low)+1:) = 1.0/Diag(Gpnt_Size(G_LOW)+1:) + &
!       CONJG(SCFvalues%Ve_fourier(2:Gpnt_Size(G_Low))) - Shift_Energy

  Diag = 1.0/Diag

!  Diag = 1.0/diag + SCFvalues%Ve_fourier(1) - Shift_Energy
!  Diag = 1.0/diag
  
!  DN = Gpnt_Size(G_LOW)-1
!  Kvec = BZ%Ku(1:3,Mem_Kpnt)
!  Do gi=1, Gpnt_Size(G_LOW)
!     G = Kvec + Gpnt(1:3,gi)
!     G2 = DOT_PRODUCT(G,G)
!     Diag(gi) = SCFvalues%Ve_Fourier(gi)-G2-Shift_Energy
!  End Do
!  Do gi=2, Gpnt_Size(G_LOW)
!     G = Kvec - Gpnt(1:3,gi)
!     G2 = DOT_PRODUCT(G,G)
!     Diag(gi+DN) = SCFvalues%Ve_Fourier(gi)-G2-Shift_Energy
!  End Do

    
  Q(1:N) = Q(1:N)*Diag
  Call Freebuffer(Diag)

  Return
End Subroutine

!******************************************************************************
! Psolve_Hermitian
!******************************************************************************

subroutine psolve_hermitian (n, x, nq, q, kz, invqkz, ldqkz, ipiv, f)
  implicit none
  integer n, nq, ldqkz, ipiv(*)
  complex x(*), q(n,*), kz(n,*), invqkz(ldqkz,*), f(*)
  
  integer info
  Complex zero, one
  parameter (zero=(0.0d0,0.0d0), one=(1.0d0,0.0d0))
  
  !.. Executable Statements ..
  call precon( n, x )
  call zgemv ('c', n, nq, one, q, n, x, 1, zero, f, 1)
  call zgetrs('n', nq, 1, invqkz, ldqkz, ipiv, f, ldqkz, info)
  call zgemv ('n', n, nq, -one, kz, n, f, 1, one, x, 1)

  Return
End Subroutine    


!******************************************************************************
!  
!  gmres_hermitian - Hermitian GMRES solver
!
!******************************************************************************

subroutine gmres_hermitian (n, x, r, mxm, eps, mxmv, &
    alpha, beta, k, kz, q, invqkz, ldqkz, ipiv, f, v, tp)
!
!     Programmer: Diederik R. Fokkema
!
      implicit none

      integer :: mxm, mxmv, n, k, ldqkz, ipiv(*)
      Real    :: eps
      Complex :: x(*), r(*), kz(n,*), q(n,*), v(n,*), &
                 alpha, beta, invqkz(ldqkz,*), f(*), tp(*)

!tex@ \begin{manpage}{ZGMRES} 
!tex@
!tex@ \subtitle{ZGMRES} 
!tex@    ZGMRES -- Generalized Minimal Residual
!tex@    iterative method for solving linear systems $\beta A-\alpha B = -r$.
!tex@    This subroutine in specilized for use in JDQZ.
!tex@ 
!tex@ \subtitle{Declaration}
!tex@ \function{subroutine zgmres (n, x, r, mxm, eps, mxmv, a, ka, b, kb,
!tex@   alpha, beta, k, kz, mqkz, zmqkz, ldvs, q,
!tex@   lu, klu, dlu, v)}
!tex@
!tex@ \subtitle{Parameters}
!tex@    \variable{integer n} 
!tex@       On entry, n specifies the dimension of the matrix A.
!tex@       
!tex@    \variable{x} 
!tex@       double complex array of size n. 
!tex@       On exit, x is overwritten by the approximate solution.
!tex@
!tex@    \variable{r}
!tex@       double complex array of size n. Before entry, the array r 
!tex@       must contain the righthandside of the linear problem Ax=r. 
!tex@       Changed on exit.
!tex@ 
!tex@    \variable{integer mxm} 
!tex@       On entry, mxm specifies the degree of the Minimal Residual
!tex@       polynomial.
!tex@
!tex@    \variable{{double precision} eps}
!tex@       On entry, eps specifies the stop tolerance. On exit, eps contains
!tex@       the relative norm of the last residual.
!tex@
!tex@    \variable{integer mxmv}
!tex@       On Entry, mxmv specifies the maximum number of matrix 
!tex@       multiplications. On exit, mxmv contains the number of matrix
!tex@       multiplications performed by the method.
!tex@
!tex@    \variable{{double complex} zalpha}
!tex@       On entry, zalpha specifies $\alpha$. Unchanged on exit.
!tex@ 
!tex@    \variable{{double complex} zbeta}
!tex@       On entry, zbeta specifies $\beta$. Unchanged on exit.
!tex@ 
!tex@    \variable{integer k}
!tex@       On entry, k specifies the number of columns of the arrays
!tex@       kz and q.
!tex@
!tex@    \variable{z}
!tex@       double complex array z, of size (n,k). On entry the array z
!tex@       must contain the preconditioned matrix Z.
!tex@
!tex@    \variable{mqkz}
!tex@       double complex array mqkz, of size (ldvs,k). On entry the array 
!tex@       mqkz must contain the matrix Q'*KZ.
!tex@
!tex@    \variable{zmqkz}
!tex@       double complex array zmqkz, of size (ldvs,k). Workspace. Used to
!tex@       copy mqkz.
!tex@
!tex@    \variable{q}
!tex@       double complex array q, of size (n,k). On entry the array q
!tex@       must contain the preconditioned matrix Q.
!tex@
!tex@    \variable{v}
!tex@       double complex array of size (n,mxm+1). Workspace.
!tex@
!tex@ \subtitle{Description}
!tex@    ***
!tex@
!tex@ \subtitle{See Also}
!tex@    ***
!tex@
!tex@ \subtitle{References}
!tex@    ***
!tex@ 
!tex@ \subtitle{Bugs}
!tex@    ***
!tex@
!tex@ \subtitle{Author}
!tex@     Diederik R.\ Fokkema
!tex@
!tex@ \end{manpage}
!tex@ \begin{verbatim}
!tex@    % actual code
!tex@ \end{verbatim}
!tex@

!     .. Local ..

      logical restrt, loop
      integer maxm
      parameter (maxm = 100)

      integer i, m, m1, nmv
      double precision rnrm0, rnrm, eps1, c(maxm)
      double complex hh(maxm,maxm-1), rs(maxm), s(maxm), y(maxm), rcs
      double complex zero, one
      parameter (zero = (0.0d0,0.0d0), one = (1.0d0,0.0d0))

      double precision  dznrm2
      double complex ztmp, zdotc


      if ( mxm .gt. maxm-1 ) &
           call error ('mxm larger than maxm (zgmres)')

      !--- Initialize first residue
      call zzeros (n, x)
      call zscal (n, -one, r, 1)
      call psolve (n, r, k, q, kz, invqkz, ldqkz, ipiv, f)

      !--- initialize loop
      rnrm0  = dznrm2 (n,r,1)
      rnrm = rnrm0
      eps1  = eps * rnrm

      nmv = 0

      call zcopy (n, r, 1, v(1,1), 1)

      !--- outer loop

 1000 restrt = ( nmv .lt. mxmv .and. rnrm .gt. eps1 )
      if ( restrt ) then
         ztmp = one / rnrm
         call zscal (n, ztmp, v(1,1), 1)
         rs(1) = rnrm

         !--- inner loop
         m = 0
 2000    loop = (nmv.lt.mxmv .and. m.lt.mxm .and. rnrm.gt.eps1)
         if (loop) then
            m  = m + 1
            m1 = m + 1
            call jdqzmv (n, v(1,m), v(1,m1), tp, alpha, beta)
            call psolve (n, v(1,m1), k, q, kz, invqkz, ldqkz, ipiv, f)
            nmv = nmv+1 
            do i = 1,m
               ztmp = zdotc (n, v(1,i), 1, v(1,m1), 1)
               hh(i,m) = ztmp
               call zaxpy (n, (-ztmp), v(1,i), 1, v(1,m1), 1)
            enddo
            ztmp = dznrm2( n, v(1,m1), 1 )
            hh(m1,m) = ztmp
            call zscal ( n, (one/ztmp), v(1,m1), 1 )
            do i = 1,m-1
               call zrot (1, hh(i,m), 1, hh(i+1,m), 1, c(i), s(i))
            enddo
            call zlartg (hh(m,m), hh(m1,m), c(m), s(m), rcs )
            hh(m,m) = rcs
            hh(m1,m) = zero
            rs(m1) = zero
            call zrot (1, rs(m), 1, rs(m1), 1, c(m), s(m))
            rnrm = abs (rs(m1))
            goto 2000
         endif

         !--- compute approximate solution x
         call zcopy ( m, rs, 1, y, 1 )
         call ztrsv ( 'u', 'n', 'n', m, hh, maxm, y, 1 )
         call zgemv ( 'n', n, m, one, v, n, y, 1, one, x, 1 )

         !--- compute residual for restart
         call jdqzmv (n, x, v(1,2), tp, alpha, beta)
         call psolve (n, v(1,2), k, q, kz, invqkz, ldqkz, ipiv, f)
         call zcopy (n, r, 1, v(1,1), 1)
         call zaxpy (n, -one, v(1,2), 1, v(1,1), 1)
         rnrm = dznrm2 (n, v(1,1), 1)

         goto 1000
      endif

      eps = rnrm/rnrm0
      mxmv = nmv

      return
end Subroutine

!******************************************************************************
!  CGSTABL_Hermitian - Hermitian BiCGSTAB
!******************************************************************************

Subroutine cgstabl_Hermitian (n, x, r, l, eps, mxmv, &
     zalpha, zbeta, nk, kz, qq, invqkz, ldqkz, jpiv, f, work, lwork)

  !** Programmer: Diederik R. Fokkema

      implicit none

      integer l, n, mxmv, nk, lwork, ldqkz, jpiv(*)
      real  eps
      complex x(*), r(*), kz(n,*), qq(n,*), work(n,*), &
           zalpha, zbeta, invqkz(ldqkz,*), f(*)


      !.. Local ..
      integer mxl
      parameter (mxl = 32)

      integer i, j, k, m, ipiv(mxl), nmv, info

      integer u, q, w, rr, xp, bp
      logical rcomp, xadd

      Real :: maxnorm, delta, bpnorm, tr0norm, trlnorm
      complex varrho, hatgamma

      Real :: rnrm, rnrm0, eps1, dznrm2
      complex alpha, beta, omega, gamma, rho0, rho1, &
           yr(mxl), ys(mxl), z(mxl,mxl), ztmp

      Complex ::  zero, one
      parameter (zero = (0.0d0,0.0d0), one = (1.0d0,0.0d0))

      complex zdotc

      !.. Executable statements ..
      if (mxmv.eq.0) return

      if (l.gt.mxl) &
           call error ('l exceeds mxl (zcgstabl)')

      u  = 1
      q  = u + (l+1)
      rr = q + (l+1)
      xp = rr + 1
      bp = xp + 1
      w = bp + 1

      if (w.gt.lwork) &
           call error ('workspace too small (zcgstabl)')

      !--- set x to zero and compute first residue
      call zzeros (n, x)
      call zscal (n, -one, r, 1)
      call zcopy (n, r, 1, work(1,rr), 1)
      call psolve (n, r, nk, qq, kz, invqkz, ldqkz,jpiv, f)

      !--- initialize loop
      nmv = 0

      rnrm0 = dznrm2 (n, r, 1)
      rnrm = rnrm0
      eps1 = eps * rnrm0

      call zcopy (n, x, 1, work(1,xp), 1)
      call zcopy (n, r, 1, work(1,bp), 1)
      maxnorm = 0d0
      bpnorm = rnrm
      rcomp = .false.
      xadd = .false.
      delta = 1d-2

      m = 0
      rho0 = one
      alpha = one
      beta = zero
      omega = one

      call zzeros (n*(l+1), work(1,u))
      call zzeros (n*(l+1), work(1,q))
      call zcopy (n, r, 1, work(1,q), 1)

      !--- loop
 1000 continue
      m = m + l

      !--- BiCG part
      rho0 = -omega * rho0
      do k=1,l
         rho1 = zdotc (n, work(1,rr), 1, work(1,q+k-1), 1)
         beta = alpha * (rho1 / rho0)
         rho0 = rho1
         beta = beta
         do j=0,k-1
            call zxpay (n, work(1,q+j), 1, (-beta), work(1,u+j), 1)
         enddo
         call jdqzmv (n, work(1,u+k-1), work(1,u+k), work(1,w), zalpha, zbeta)
         call psolve (n, work(1,u+k), nk, qq, kz, invqkz, ldqkz, jpiv, f)
         gamma = zdotc (n, work(1,rr), 1, work(1,u+k), 1)
         alpha = rho0 / gamma
         do j=0,k-1
            call zaxpy (n, (-alpha), work(1,u+j+1), 1, work(1,q+j), 1)
         enddo
         call jdqzmv (n, work(1,q+k-1), work(1,q+k), work(1,w), zalpha, zbeta)
         call psolve (n, work(1,q+k), nk, qq, kz, invqkz, ldqkz, jpiv, f)
         call zaxpy (n, alpha, work(1,u), 1, x, 1)
         rnrm = dznrm2 (n, work(1,q), 1)
         maxnorm = max (maxnorm, rnrm)
         nmv = nmv+2
      enddo

      !--- MR part + Maintaining the convergence
      do i=1,l-1
         do j=1,i
            ztmp = zdotc (n, work(1,q+i), 1, work(1,q+j), 1)
            z(i,j) = ztmp
            z(j,i) = dconjg(ztmp)
         enddo
         yr(i) = zdotc (n, work(1,q+i), 1, work(1,q), 1)
         ys(i) = zdotc (n, work(1,q+i), 1, work(1,q+l), 1)
      enddo
      call zgetrf (l-1, l-1, z, mxl, ipiv, info)
      call zgetrs ('n', l-1, 1, z, mxl, ipiv, yr, mxl, info)
      call zgetrs ('n', l-1, 1, z, mxl, ipiv, ys, mxl, info)
      call zcopy (n, work(1,q), 1, r, 1)
      call zgemv ('n', n, l-1, (-one), work(1,q+1), n, yr, 1, one,r, 1)
      call zgemv ('n', n, l-1, (-one), work(1,q+1), n,ys,1,one,work(1,q+l), 1)
      tr0norm = dznrm2 (n, r, 1)
      trlnorm = dznrm2 (n, work(1,q+l), 1)
      varrho = zdotc (n, work(1,q+l), 1, r, 1) / (tr0norm*trlnorm)
      hatgamma = varrho/abs(varrho) * max(abs(varrho),7d-1)
      hatgamma = (tr0norm/trlnorm)*hatgamma
      yr(l) = zero
      ys(l) = -one
      call zaxpy (l, (-hatgamma), ys, 1, yr, 1)

      omega = yr(l)
      call zgemv ('n', n, l, one, work(1,q), n, yr, 1, one, x, 1)
      call zgemv ('n', n, l, (-one), work(1,u+1), n, yr, 1, one,work(1,u), 1)
      call zaxpy (n, (-hatgamma), work(1,q+l), 1, r, 1)
      call zcopy (n, r, 1, work(1,q), 1)

      !--- reliable update
      rnrm = dznrm2 (n, work(1,q), 1)
      maxnorm = max (maxnorm, rnrm)
      xadd = (rnrm.lt.delta*rnrm0.and.rnrm0.lt.maxnorm)
      rcomp = ((rnrm.lt.delta*maxnorm.and.rnrm0.lt.maxnorm).or.xadd)

      if (rcomp) then
         call jdqzmv (n, x, work(1,q), work(1,w),zalpha, zbeta)
         call psolve (n, work(1,q), nk, qq, kz, invqkz, ldqkz, jpiv, f)
         call zxpay (n, work(1,bp), 1, -one, work(1,q), 1)
         maxnorm = rnrm
         if (xadd) then
            call zaxpy (n, one, x, 1, work(1,xp), 1)
            call zzeros (n, x)
            call zcopy (n, work(1,q), 1, work(1,bp), 1)
            bpnorm = rnrm
         endif
      endif
         
      if (nmv.lt.mxmv .and. rnrm.gt.eps1) goto 1000

      call zaxpy (n, one, work(1,xp), 1, x, 1)

      mxmv = nmv
      eps = rnrm/rnrm0

      return
end Subroutine 

!******************************************************************************
! hmgs - hermitian mgs
!******************************************************************************

subroutine hmgs(n, k, q, z, x)
  Integer :: n, k, job
  Complex :: q(n,*), z(n,*), x(n) 
  
  Integer :: i

  do i=1, k
     x(1:N) = x(1:N) - q(1:N,i)*DOT_PRODUCT(z(1:N,i), x(1:N))
  End Do

  Return
End Subroutine

!******************************************************************************
! mgs_weight - returns the orthog weights and orthogs
!******************************************************************************

subroutine mgs_weight(n, k, q, z, x, Weight)
  Integer :: n, k
  Complex :: q(n,*), z(n,*), x(n), Weight(k)
  
  Integer :: i

  do i=1, k
     Weight(i) = DOT_PRODUCT(z(1:N,i), x(1:N))
     x(1:N) = x(1:N) - q(1:N,i)*Weight(i)
  End Do

  Return
End Subroutine


!******************************************************************************
!  QZSort_Filter - Rearrages the schur matrices based on "order" and
!     also filters the eigenvalues less than Emin.
!
!     Originally Coded by Diederik Fokkema and 
!     modified by Alan Tackett to add the Emin filtering
!******************************************************************************
Subroutine QZSort_Filter (Emin, ta, tb, k, s, t, z, q, ldz, alpha, beta, order)
  implicit none
  integer k, ldz, order
  complex :: ta, tb, s(ldz,*), t(ldz,*), z(ldz,*), q(ldz,*), alpha(*), beta(*)
  Real    :: Emin

  integer :: i, j, select
  Real    :: eigen

  do i = 1,k
     do j = 1,k
        alpha(j) = s(j,j)
        beta(j) = t(j,j)
        eigen = alpha(j)/beta(j)
        if (eigen < Emin) then
           alpha(j) = 1E20
           beta(j)  = 1
        End if
     end do
     j = select (k-i+1, ta, tb, alpha(i), beta(i), order) + i-1
     call myexc (k, s, t, z, q, ldz, j, i)
  end do

  Return
end Subroutine
